/* .MiK. SoftWare

  Multilingual Terminal Emulator "MTR".

 Copyright (C) 2000 Mike I. Kozin. ALL RIGHT RESERVED
 
 On Base "MuTerminal"
 Copyright (C) 1998,99 Kazuho Okui and Takashi Murai. ALL RIGHT RESERVED
  
 This file is part of MTR.

 MTR is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 CodeConv.cpp,v 0.1 26-Sep-2000
***************************************************************************/

#include <support/UTF8.h>
#include <string.h>

#include "CodeConv.h"

extern char utf8_width_table[]; /* define UTF8WidthTbl.c */

//////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////
CodeConv::CodeConv (void)
{

}
//////////////////////////////////////////////////////////////////////
// Destructor
//////////////////////////////////////////////////////////////////////
CodeConv::~CodeConv (void)
{

}
//////////////////////////////////////////////////////////////////////
// UTF8GetFontWidth (const uchar *string)
// get font width in coding.
//////////////////////////////////////////////////////////////////////
int
CodeConv::UTF8GetFontWidth (const uchar *string)
{
  uchar  width, point;
  ushort unicode, offset;
  
  offset = unicode = UTF8toUnicode (string);
  width = utf8_width_table [unicode >> 3];
  offset = offset & 0x07;

  point = 0x80 >> offset;
  
  return ((width & point) > 0 ) ? 2 : 1;
}

//////////////////////////////////////////////////////////////////////
// ConvertFromInternal (const uchar *src, uchar *dst, int coding)
//	 Convert internal coding from src coding.
//////////////////////////////////////////////////////////////////////
int
CodeConv::ConvertFromInternal (const uchar *src, uchar *dst, int coding)
{
  const uchar *src_p = src;
  long srclen = 0, dstlen = 0;
  long state = 0;
  int theCoding;

  if (coding == M_UTF8) return 0;

  theCoding = coding_translation_table[coding];
  
  while (*src_p++) srclen++;

  dstlen = srclen * 256;

  convert_from_utf8 (theCoding,
		     (char *)src, &srclen,
		     (char *)dst, &dstlen,
		     &state, '?');

  dst[dstlen] = '\0';
  
  return dstlen;

}
//////////////////////////////////////////////////////////////////////
// ConvertToInternal (const uchar *src, uchar *dst, int coding)
//	Convert internal coding to src coding.
//////////////////////////////////////////////////////////////////////
void
CodeConv::ConvertToInternal (const uchar *src, uchar *dst, int coding)
{
  const uchar *src_p = src;
  long srclen = 0, dstlen = 4;
  long state = 0;
  int theCoding;
  

  if (coding == M_UTF8) return;

  theCoding = coding_translation_table[coding];

  while (*src_p++) srclen++;
  
  convert_to_utf8 (theCoding,
		   (char *)src, &srclen,
		   (char *)dst, &dstlen,
		   &state, '?');

  dst[dstlen] = '\0';

  return;
}


/*
 * Private member functions.
 */
//////////////////////////////////////////////////////////////////////
// UTF8toUnicode (uchar *utf8)
// 
//////////////////////////////////////////////////////////////////////
unsigned short
CodeConv::UTF8toUnicode (const uchar *utf8)
{
  uchar tmp;
  unsigned short unicode = 0x0000;
  
  tmp = *utf8 & 0xe0;
  
  if (tmp < 0x80) {
    unicode = (unsigned short) *utf8;
  }
  else if (tmp < 0xe0) {
    unicode = (unsigned short) *utf8++ & 0x1f;
    unicode = unicode << 6;
    unicode = unicode | ((unsigned short) *utf8 & 0x3f);
  }
  else if (tmp < 0xf0) {
    unicode = (unsigned short) *utf8++ & 0x0f;
    unicode = unicode << 6;
    unicode = unicode | ((unsigned short) *utf8++ & 0x3f);
    unicode = unicode << 6;
    unicode = unicode | ((unsigned short) *utf8 & 0x3f);
  }

  return unicode;
}
